#include "renderer2D.hpp"

#include "exceptions.hpp"
#include "vulkan_allocator.hpp"
#include "vulkan_instance.hpp"
#include "texture_manager.hpp"

#include <cstdint>
#include <cstring>
#include <stdexcept>
#include <thread>
#include <vulkan/vulkan_enums.hpp>

namespace gz::vlk {

//
// INIT & CLEANUP
//
    Renderer2D::Renderer2D(VulkanInstance& instance, TextureManager& textureManager) : 
        Renderer(instance, textureManager)
    {
        LogCreateInfo logCI {
        	.logfile = "renderer2D.log",
        	.storeLog = false,
        	.prefix = "2D-Renderer",
        	.prefixColor = Color::LI_MAGENTA,
        	.timeColor = settings::VULKAN_MESSAGE_TIME_COLOR,
        };
        rLog = Log(std::move(logCI));

        vk.registerCleanupCallback(std::bind(&Renderer2D::cleanup, this));
        vk.registerSwapChainRecreateCallback(std::bind(&Renderer2D::swapChainRecreateCallback, this));
        vk.createCommandBuffers(commandBuffers);
        const size_t vertexCount = 500;
        const size_t indexCount = 1000;
        vk.createVertexBuffer<Vertex2D>(vertexCount, vertexBuffer, vertexBufferMemory, vertexBufferSize);
        vk.createIndexBuffer<uint32_t>(indexCount, indexBuffer, indexBufferMemory, indexBufferSize);
        initSwapChainDependantResources();
        VulkanInstance::registerObjectUsingVulkan("Renderer2D",
                &pipelines[R2Render].pipeline, &renderPass, &vertexBuffer, &vertexBufferMemory.memory, &indexBuffer, &indexBufferMemory.memory,
                &framebuffers, &renderImages, &renderImageViews, &commandBuffers);  // &renderImageMemory,
        rLog("Created Renderer2D");
    }


    void Renderer2D::cleanup() {
        rLog.log0("cleanup: cleaning up");
        // UPDATE DOC ON CHANGES!

        /* vk.destroyCommandBuffers(commandBuffers); */
        cleanupSwapChainDependantResources();
        cleanup_();
    }


//
// SWAPCHAIN DEPENDANT
//
    void Renderer2D::initSwapChainDependantResources() {
        createRenderPass();
        createImageResources();
        vk.createFramebuffers(framebuffers, renderImageViews, renderPass);
        std::vector<vk::DescriptorSetLayout> descriptorSetLayouts = { textureManager.getDescriptorSetLayout() };

        // specialization constant for sampler2D array length = atlas count
        vk::SpecializationMapEntry specME {
            .constantID = 0,
            .offset = 0,
            .size = sizeof(uint32_t),
        };
        uint32_t atlasCount = textureManager.getAtlasCount();
        vk::SpecializationInfo specI;
        specI.setData<uint32_t>(atlasCount);
        specI.setMapEntries(specME);

        // VERTEX 2D PIPELINE
        auto vertShaderCode = readBinaryFile("shaders/shader2D.vert.spv");
        auto fragShaderCode = readBinaryFile("shaders/shader2D.frag.spv");

        vk::ShaderModule vertShaderModule = vk.createShaderModule(vertShaderCode);
        vk::ShaderModule fragShaderModule = vk.createShaderModule(fragShaderCode);

        std::vector<vk::PipelineShaderStageCreateInfo> shaderStages;

        shaderStages.emplace_back(vk::PipelineShaderStageCreateInfo {
        	.stage = vk::ShaderStageFlagBits::eVertex,
        	.module = vertShaderModule,
        	.pName = "main",
        	.pSpecializationInfo = nullptr,
        });
        shaderStages.emplace_back(vk::PipelineShaderStageCreateInfo {
        	.stage = vk::ShaderStageFlagBits::eFragment,
        	.module = fragShaderModule,
        	.pName = "main",
        	.pSpecializationInfo = nullptr,
        });

        vk::GraphicsPipelineCreateInfo pipelineCI {
            .renderPass = renderPass,
            .subpass = 0,
        };
        pipelineCI.setStages(shaderStages);

        vk.createGraphicsPipeline<Vertex3D>(std::move(pipelineCI), &descriptorSetLayouts, true, pipelines[R2Render]);

        vk.getDevice().destroyShaderModule(vertShaderModule, nullptr);
        vk.getDevice().destroyShaderModule(fragShaderModule, nullptr);
    }


    void Renderer2D::cleanupSwapChainDependantResources() {
        // destroy pipelines
        pipelines.clear(vk.getDevice());

        vk.destroyFramebuffers(framebuffers);

#ifndef RENDERER_2D_RENDER_TO_SWAP_CHAIN
        for (size_t i = 0; i < renderImages.size(); i++) {
            vk.destroyImageView(renderImageViews[i]);
            vk.destroyImage(renderImages[i], renderImageMemory[i]);
        }
#endif
        vk.getDevice().destroyRenderPass(renderPass, nullptr);
    }


    void Renderer2D::swapChainRecreateCallback() {
        cleanupSwapChainDependantResources();
        initSwapChainDependantResources();
    }


//
// IMAGES
//
    void Renderer2D::createImageResources() {
        renderImageViews.resize(vk.getScImages().size());
#ifndef RENDERER_2D_RENDER_TO_SWAP_CHAIN
        renderImages.resize(vk.getScImages().size());
        renderImageMemory.resize(vk.getScImages().size());
        vk::ImageUsageFlags usage= vk::ImageUsageFlagBits::eColorAttachment | vk::ImageUsageFlagBits::eTransferSrc;
        for (size_t i = 0; i < renderImages.size(); i++) {
            vk.createImage(vk.getScExtent().width, vk.getScExtent().height, vk.getScImageFormat(), vk::ImageTiling::eOptimal, usage, vk::MemoryPropertyFlagBits::eDeviceLocal, renderImages[i], renderImageMemory[i]);
            vk.createImageView(vk.getScImageFormat(), renderImages[i], renderImageViews[i], vk::ImageAspectFlagBits::eColor);
        }
        vk.createTextureSampler(imageSampler);
        rLog.log0("createImageResources: created images, views and sampler");
#else
        for (size_t i = 0; i < renderImages.size(); i++) {
            vk.createImageView(vk.getScImageFormat(), vk.getScImages()[i], renderImageViews[i], vk::ImageAspectFlagBits::eColor);
        }
        rLog.log0("createImageResources: created image views");
#endif
    }

//
// RENDER PASS
//
    void Renderer2D::createRenderPass() {
        vk::AttachmentDescription2 colorBlendAttachment {
            .format = vk.getScImageFormat(),
            .samples = vk::SampleCountFlagBits::e1,
            .loadOp = vk::AttachmentLoadOp::eClear,
            .storeOp = vk::AttachmentStoreOp::eStore,
            .stencilLoadOp = vk::AttachmentLoadOp::eDontCare,
            .stencilStoreOp = vk::AttachmentStoreOp::eDontCare,
            .initialLayout = vk::ImageLayout::eUndefined,
            .finalLayout = vk::ImageLayout::eTransferSrcOptimal,
        };

        vk::AttachmentReference2 colorAttachmentRef {
            .attachment = 0,
            .layout = vk::ImageLayout::eColorAttachmentOptimal,
        };

        /* vk::AttachmentDescription depthAttachment { */
        /* depthAttachment.format = findDepthFormat(); */
        /* depthAttachment.samples = vk::SampleCountFlagBits::e1; */
        /* depthAttachment.loadOp = vk::AttachmentLoadOp::eClear; */
        /* depthAttachment.storeOp = vk::AttachmentStoreOp::eDontCare; */
        /* depthAttachment.stencilLoadOp = vk::AttachmentLoadOp::eDontCare; */
        /* depthAttachment.stencilStoreOp = vk::AttachmentStoreOp::eDontCare; */
        /* depthAttachment.initialLayout = vk::ImageLayout::eUndefined; */
        /* depthAttachment.finalLayout = vk::ImageLayout::eDepthStencilAttachmentOptimal; */

        /* vk::AttachmentReference depthAttachmentRef { */
        /* depthAttachmentRef.attachment = 1; */
        /* depthAttachmentRef.layout = vk::ImageLayout::eDepthStencilAttachmentOptimal; */

        // SUBPASS 1: DRAW STUFF
        vk::SubpassDescription2 drawSubpass {
            .pipelineBindPoint = vk::PipelineBindPoint::eGraphics,
            .colorAttachmentCount = 1,
            .pColorAttachments = &colorAttachmentRef,
            /* .pDepthStencilAttachment = &depthAttachmentRef, */
        };

        vk::SubpassDependency2 dsColorAttachmentSD {
            .srcSubpass = VK_SUBPASS_EXTERNAL,
            .dstSubpass = 0,
            .srcStageMask = vk::PipelineStageFlagBits::eColorAttachmentOutput,
            .dstStageMask = vk::PipelineStageFlagBits::eColorAttachmentOutput,
            .srcAccessMask = NO_ACC_FLAGS,
            .dstAccessMask = vk::AccessFlagBits::eColorAttachmentWrite,
        };

        // dependecy for the image layout transition to transfer dst
        /* vk::SubpassDependency2 dsLayoutTransitionSD { */
        /*     .srcSubpass = 0, */
        /*     .dstSubpass = 1, */
        /*     .srcStageMask = vk::PipelineStageFlagBits::eColorAttachmentOutput, */
        /*     .dstStageMask = vk::PipelineStageFlagBits::eTransfer, */
        /*     .srcAccessMask = vk::AccessFlagBits::eColorAttachmentWrite, */
        /*     .dstAccessMask = vk::AccessFlagBits::eTransferRead, */
        /*     .dependencyFlags = vk::DependencyFlagBits::eByRegion, */
        /* }; */

        // SUBPASS 2: BLIT ONTO SWAP CHAIN IMAGE
        vk::SubpassDescription2 blitSubpass {
            .pipelineBindPoint = vk::PipelineBindPoint::eGraphics,
            .colorAttachmentCount = 1,
            .pColorAttachments = &colorAttachmentRef,
        };
        
        vk::SubpassDependency2 bsColorAttachmentSD {
            .srcSubpass = 0,
            .dstSubpass = 1,
            .srcStageMask = vk::PipelineStageFlagBits::eColorAttachmentOutput,
            .dstStageMask = vk::PipelineStageFlagBits::eColorAttachmentOutput,
            .srcAccessMask = vk::AccessFlagBits::eColorAttachmentWrite,
            .dstAccessMask = vk::AccessFlagBits::eColorAttachmentRead,
        };

        /* std::array<vk::AttachmentDescription, 2> attachments = { colorBlendAttachment, depthAttachment }; */
        std::vector<vk::AttachmentDescription2> attachments = { colorBlendAttachment };
        std::vector<vk::SubpassDescription2> subpasses { drawSubpass, blitSubpass };
        std::vector<vk::SubpassDependency2> dependencies = { dsColorAttachmentSD, bsColorAttachmentSD }; //dsLayoutTransitionSD };
        vk::RenderPassCreateInfo2 renderPassCI;
        renderPassCI.setAttachments(attachments);
        renderPassCI.setSubpasses(subpasses);
        renderPassCI.setDependencies(dependencies);

        vk::Result result = vk.getDevice().createRenderPass2(&renderPassCI, nullptr, &renderPass); 
        if (result != vk::Result::eSuccess) {
            throw getVkException(result, "Could not create render pass", "Renderer2D::createRenderPass");
        }
        rLog("createRenderPass: Created render pass.");
    }


//
// RENDERING
//
    void Renderer2D::recordCommandBuffer(uint32_t imageIndex, uint32_t currentFrame) {
        vk::CommandBufferBeginInfo commandBufferBI {
        	/* .flags = 0, */
        	/* .pInheritanceInfo = nullptr, */
        };
        vk::Result result = commandBuffers[currentFrame].begin(&commandBufferBI);
        if (result != vk::Result::eSuccess) {
            throw getVkException(result, "Failed to begin 2D command buffer", "Renderer2D::recordCommandBuffer");
        }

        // clear
        std::array<vk::ClearValue, 1> clearValues{};
        vk::ClearColorValue clearColor{};
        clearColor.setFloat32({ 0.0f, 0.0f, 0.1f, 0.5f });
        clearValues[0].setColor(std::move(clearColor));
        /* clearValues[1].depthStencil = {1.0f, 0}; */
        vk::RenderPassBeginInfo renderPassBI {
        	.renderPass = renderPass,
        	.framebuffer = framebuffers[imageIndex],
            .renderArea { .offset = { 0, 0 }, .extent = vk.getScExtent(), }
        };
        renderPassBI.setClearValues(clearValues);

        commandBuffers[currentFrame].beginRenderPass(&renderPassBI, vk::SubpassContents::eInline);

        commandBuffers[currentFrame].bindPipeline(vk::PipelineBindPoint::eGraphics, pipelines[R2Render].pipeline);
        vk::Buffer vertexBuffers[] = { vertexBuffer };
        vk::DeviceSize offsets[] = {0};
        uint32_t bindingCount = 1;
        uint32_t binding = 0;
        commandBuffers[currentFrame].bindVertexBuffers(binding, bindingCount, vertexBuffers, offsets);
        // TODO use correct index type!
        commandBuffers[currentFrame].bindIndexBuffer(indexBuffer, NO_OFFSET, vk::IndexType::eUint32);

        uint32_t descriptorCount = 1;
        uint32_t firstSet = 0;
        uint32_t dynamicOffsetCount = 0;
        uint32_t* dynamicOffsets = nullptr;
        commandBuffers[currentFrame].bindDescriptorSets(vk::PipelineBindPoint::eGraphics, pipelines[R2Render].layout, firstSet, descriptorCount, &textureManager.getDescriptorSet(),  dynamicOffsetCount, dynamicOffsets);

        int instanceCount = 1;
        int firstIndex = 0;
        int firstInstance = 0;
        commandBuffers[currentFrame].drawIndexed(static_cast<uint32_t>(shapesIndicesCount), instanceCount, firstIndex, NO_OFFSET, firstInstance);
        commandBuffers[currentFrame].endRenderPass();

        /* vk.copyImageToImage(commandBuffers[currentFrame], images[imageIndex], vk.getScImages()[imageIndex], vk.getScExtent()); */
        result = commandBuffers[currentFrame].end();
        if (result != vk::Result::eSuccess) {
            rLog.error("Failed to record 2D - command buffer", "VkResult:", result);
            throw getVkException(result, "Failed to record 2D - command buffer", "Renderer2D::recordCommandBufferWithTexture");
        }
        vk.submitThisFrame(commandBuffers[currentFrame]);
    }


    void Renderer2D::fillVertexBufferWithShapes() {
        rLog("fillVertexBufferWithShapes");
        if (vertexBufferSize < shapesVerticesCount * sizeof(Vertex2D)) {
            throw VkException("vertex buffer too small! vertexBufferSize: " + std::to_string(vertexBufferSize) + ", required size: " + std::to_string(shapesVerticesCount), "fillVertexBufferWithShapes");
        }

        // create staging buffer
        vk::Buffer stagingBuffer;
        MemoryInfo stagingBufferMemory;
        vk.createBuffer(vertexBufferSize, vk::BufferUsageFlagBits::eTransferSrc, vk::MemoryPropertyFlagBits::eHostVisible | vk::MemoryPropertyFlagBits::eHostCoherent, stagingBuffer, stagingBufferMemory);

        // fill staging buffer
        void* data;
        vk::Result result = vk.getDevice().mapMemory(stagingBufferMemory.memory, stagingBufferMemory.offset, vertexBufferSize, NO_MEM_FLAGS, &data);
        if (result != vk::Result::eSuccess) {
            throw getVkException(result, "Failed to map staging buffer", "Renderer3D::fillVertexBufferWithShapes");
        }
        Vertex2D* vdata = reinterpret_cast<Vertex2D*>(data);
        size_t offset = 0;
        for (auto it = shapes.begin(); it != shapes.end(); it++) {
            rLog.log0("fillVertexBufferWithShapes: copying vertex buffer nr", it - shapes.begin(), "-", it->getVertices(), "to address:", long(vdata + offset), " offset:", offset);
            memcpy(vdata+offset, it->getVertices().data(), it->getVertices().size() * sizeof(Vertex2D));
            offset += it->getVertices().size();
        }
        vk.getDevice().unmapMemory(stagingBufferMemory.memory);
        // fill vertex buffer 
        vk.copyBuffer(stagingBuffer, vertexBuffer, vertexBufferSize);
        vk.destroyBuffer(stagingBuffer, stagingBufferMemory);
    }
    void Renderer2D::fillIndexBufferWithShapes() {
        rLog("fillIndexBufferWithShapes");
        if (indexBufferSize < shapesIndicesCount * sizeof(uint32_t)) {
            throw VkException("index buffer too small! indexBufferSize: " + std::to_string(vertexBufferSize) + ", required size: " + std::to_string(shapesVerticesCount), "fillVertexBufferWithShapes");
        }

        // create staging buffer
        vk::Buffer stagingBuffer;
        MemoryInfo stagingBufferMemory;
        vk.createBuffer(indexBufferSize, vk::BufferUsageFlagBits::eTransferSrc, vk::MemoryPropertyFlagBits::eHostVisible | vk::MemoryPropertyFlagBits::eHostCoherent, stagingBuffer, stagingBufferMemory);

        // fill staging buffer
        void* data;
        vk::Result result = vk.getDevice().mapMemory(stagingBufferMemory.memory, stagingBufferMemory.offset, indexBufferSize, NO_MEM_FLAGS, &data);
        if (result != vk::Result::eSuccess) {
            throw getVkException(result, "Failed to map staging buffer", "Renderer2D::fillIndexBufferWithShapes");
        }
        uint32_t* idata = reinterpret_cast<uint32_t*>(data);
        size_t offset = 0;
        for (auto it = shapes.begin(); it != shapes.end(); it++) {
            rLog("fillIndexBufferWithShapes: copying index buffer nr", it - shapes.begin(), "-", it->getIndices(), "to address:", long(idata + offset), " offset:", offset);
            memcpy(idata+offset, it->getIndices().data(), it->getIndices().size() * sizeof(uint32_t));
            offset += it->getIndices().size();
        }
        rLog.log0("fillIndexBufferWithShapes: indices count:", shapesIndicesCount);
        vk.getDevice().unmapMemory(stagingBufferMemory.memory);

        // fill index buffer
        vk.copyBuffer(stagingBuffer, indexBuffer, indexBufferSize);
        vk.destroyBuffer(stagingBuffer, stagingBufferMemory);
        
    }


    void Renderer2D::drawShape(Shape* shape) {
        // make indices valid
        shape->setIndexOffset(shapesVerticesCount);
        shape->normalizeVertices(vk.getScExtent().width, vk.getScExtent().height);
        shape->setTextureCoordinates(textureManager);
        // object slicing here, need to call virtual setTextureCoordinates before this!
        shapes.push_back(*shape);
        shapesVerticesCount += shape->getVertices().size();
        shapesIndicesCount += shape->getIndices().size();
    }


    void Renderer2D::drawFrame(uint32_t imageIndex) {
        commandBuffers[vk.getCurrentFrame()].reset();
        /* recordCommandBuffer(imageIndex, vk.getCurrentFrame()); */
        recordCommandBuffer(imageIndex, vk.getCurrentFrame());

    }


}  // namespace gz::vlk
