#include "renderer3D.hpp"

#include "vertex.hpp"
#include "vulkan_allocator.hpp"
#include "vulkan_instance.hpp"
#include "texture_manager.hpp"
#include "exceptions.hpp"
#include "vk_enum_string.h"
#include "vulkan_settings.hpp"

#include <gz-util/file_io.hpp>

#include <cstring>
#include <chrono>
#include <vulkan/vulkan_structs.hpp>

namespace gz::vlk {
//
// INIT & CLEANUP
//
    Renderer3D::Renderer3D(VulkanInstance& instance, TextureManager& textureManager) : 
        Renderer(instance, textureManager) {
        LogCreateInfo logCI{
        	.logfile = "renderer3D.log",
        	.storeLog = false,
        	.prefix = "3D-Renderer",
        	.prefixColor = Color::LI_CYAN,
        	.timeColor = settings::VULKAN_MESSAGE_TIME_COLOR,
        };
        rLog = Log(std::move(logCI));

        vk.registerCleanupCallback(std::bind(&Renderer3D::cleanup, this));
        vk.registerSwapChainRecreateCallback(std::bind(&Renderer3D::swapChainRecreateCallback, this));

        vk.createCommandBuffers(commandBuffers);
        const size_t vertexCount = 20000;
        const size_t indexCount = 10000;
        vk.createVertexBuffer<Vertex3D>(vertexCount, vertexBuffer, vertexBufferMemory, vertexBufferSize);
        vk.createIndexBuffer<uint32_t>(indexCount, indexBuffer, indexBufferMemory, indexBufferSize);

        // TODO
        loadModel();

        createUniformBuffers();
        createDescriptorResources();
        initSwapChainDependantResources();
        VulkanInstance::registerObjectUsingVulkan("Renderer3D",
                &renderPass, &imageSampler, &vertexBuffer, &vertexBufferMemory.memory, &indexBuffer, &indexBufferMemory.memory,
                &pipelines[R3Render].pipeline, //&pipelines[blendToSC].pipeline,
                &descriptorPool, &uboDSL, &sampleFromRenderImageDSL,
                &framebuffers, &renderImages, &renderImageViews, &commandBuffers,
                &uboDS, &sampleFromRenderImageDS); // missing: &renderImageMemory, 
        rLog.log1("Created Renderer3D");
    }

    void Renderer3D::cleanup() {
        rLog.log0("cleanup: cleaning up");
        // UPDATE DOC ON CHANGES!
        for (size_t i = 0; i < vk.getMaxFramesInFlight(); i++) {
            vk.destroyBuffer(uniformBuffers[i], uniformBuffersMemory[i]);
        }
        vk.getDevice().destroyDescriptorSetLayout(uboDSL, NO_ALLOC);
        vk.getDevice().destroyDescriptorPool(descriptorPool, NO_ALLOC);

        cleanupSwapChainDependantResources();
        cleanup_();
    }


//
// SWAPCHAIN DEPENDANT
//
    void Renderer3D::initSwapChainDependantResources() {
        // UPDATE DOC ON CHANGES!
        createRenderPass();
        createDepthImage();
        createImageResources();
        vk.createFramebuffers(framebuffers, renderImageViews, renderPass, depthImageView);

        // VERTEX 3D PIPELINE
        std::vector<vk::DescriptorSetLayout> descriptorSetLayouts = { uboDSL, textureManager.getDescriptorSetLayout() };
        auto vertShaderCode = readBinaryFile("shaders/shader.vert.spv");
        auto fragShaderCode = readBinaryFile("shaders/shader.frag.spv");

        vk::ShaderModule vertShaderModule = vk.createShaderModule(vertShaderCode);
        vk::ShaderModule fragShaderModule = vk.createShaderModule(fragShaderCode);

        std::vector<vk::PipelineShaderStageCreateInfo> shaderStages;

        shaderStages.emplace_back(vk::PipelineShaderStageCreateInfo {
        	.stage = vk::ShaderStageFlagBits::eVertex,
        	.module = vertShaderModule,
        	.pName = "main",
        	.pSpecializationInfo = nullptr,
        });
        shaderStages.emplace_back(vk::PipelineShaderStageCreateInfo {
        	.stage = vk::ShaderStageFlagBits::eFragment,
        	.module = fragShaderModule,
        	.pName = "main",
        	.pSpecializationInfo = nullptr,
        });

        vk::GraphicsPipelineCreateInfo pipelineCI {
            .renderPass = renderPass,
            .subpass = 0,
        };
        pipelineCI.setStages(shaderStages);

        vk.createGraphicsPipeline<Vertex3D>(std::move(pipelineCI), &descriptorSetLayouts, true, pipelines[R3Render]);

        vk.getDevice().destroyShaderModule(vertShaderModule, nullptr);
        vk.getDevice().destroyShaderModule(fragShaderModule, nullptr);


        // BLIT TO SWAP CHAIN PIPELINE
        /* auto fragShaderCode2 = readBinaryFile("shaders/blend.frag.spv"); */
        /* vk::ShaderModule fragShaderModule2 = vk.createShaderModule(fragShaderCode); */

        /* std::vector<vk::PipelineShaderStageCreateInfo> shaderStages2; */
        /* shaderStages2.emplace_back(vk::PipelineShaderStageCreateInfo { */
        /* 	.stage = vk::ShaderStageFlagBits::eFragment, */
        /* 	.module = fragShaderModule2, */
        /* 	.pName = "main", */
        /* 	.pSpecializationInfo = nullptr, */
        /* }); */

        /* vk::GraphicsPipelineCreateInfo pipelineCI2 { */
        /*     .renderPass = renderPass, */
        /*     .subpass = 1, */
        /* }; */
        /* pipelineCI2.setStages(shaderStages2); */

        /* vk.createGraphicsPipeline<Vertex3D>(std::move(pipelineCI2), nullptr, false, pipelines[blendToSC]); */

        /* vk.getDevice().destroyShaderModule(fragShaderModule2, nullptr); */
    }


    void Renderer3D::cleanupSwapChainDependantResources() {
        // destroy pipelines
        pipelines.clear(vk.getDevice());

        vk.destroyFramebuffers(framebuffers);

        vk.destroyImageView(depthImageView);
        vk.destroyImage(depthImage, depthImageMemory);

#ifndef RENDERER_3D_RENDER_TO_SWAP_CHAIN
        for (size_t i = 0; i < renderImages.size(); i++) {
            vk.destroyImageView(renderImageViews[i]);
            vk.destroyImage(renderImages[i], renderImageMemory[i]);
        }
#endif
        vk.getDevice().destroyRenderPass(renderPass, nullptr);
    }


    void Renderer3D::swapChainRecreateCallback() {
        cleanupSwapChainDependantResources();
        initSwapChainDependantResources();
    }


//
// IMAGES
//
    void Renderer3D::createImageResources() {
        renderImageViews.resize(vk.getScImages().size());
#ifndef RENDERER_3D_RENDER_TO_SWAP_CHAIN
        renderImages.resize(vk.getScImages().size());
        renderImageMemory.resize(vk.getScImages().size());
        vk::ImageUsageFlags usage= vk::ImageUsageFlagBits::eColorAttachment | vk::ImageUsageFlagBits::eTransferSrc;
        for (size_t i = 0; i < renderImages.size(); i++) {
            vk.createImage(vk.getScExtent().width, vk.getScExtent().height, vk.getScImageFormat(), vk::ImageTiling::eOptimal, usage, vk::MemoryPropertyFlagBits::eDeviceLocal, renderImages[i], renderImageMemory[i]);
            vk.createImageView(vk.getScImageFormat(), renderImages[i], renderImageViews[i], vk::ImageAspectFlagBits::eColor);
        }
        vk.createTextureSampler(imageSampler);
        rLog.log0("createImageResources: created images, views and sampler");
#else
        for (size_t i = 0; i < renderImages.size(); i++) {
            vk.createImageView(vk.getScImageFormat(), vk.getScImages()[i], renderImageViews[i], vk::ImageAspectFlagBits::eColor);
        }
        rLog.log0("createImageResources: created image views");
#endif
    }


//
// DEPTH
//
    void Renderer3D::createDepthImage() {
        vk::MemoryPropertyFlags memoryProperties = vk::MemoryPropertyFlagBits::eDeviceLocal;
        vk::ImageUsageFlags imageUsage = vk::ImageUsageFlagBits::eDepthStencilAttachment;
        vk::ImageTiling tiling = vk::ImageTiling::eOptimal;
        vk::Format depthFormat = vk.getDepthFormat();

        vk.createImage(vk.getScExtent().width, vk.getScExtent().height, depthFormat, tiling, imageUsage, memoryProperties, depthImage, depthImageMemory);
        vk.createImageView(depthFormat, depthImage, depthImageView, vk::ImageAspectFlagBits::eDepth);

        vk.transitionImageLayout(depthImage, depthFormat, vk::ImageLayout::eUndefined, vk::ImageLayout::eDepthStencilAttachmentOptimal);
    }

//
// RENDER PASS
//
    void Renderer3D::createRenderPass() {
        vk::AttachmentDescription2 colorBlendAD {
            .format = vk.getScImageFormat(),
            .samples = vk::SampleCountFlagBits::e1,
            .loadOp = vk::AttachmentLoadOp::eClear,
            .storeOp = vk::AttachmentStoreOp::eStore,
            .stencilLoadOp = vk::AttachmentLoadOp::eDontCare,
            .stencilStoreOp = vk::AttachmentStoreOp::eDontCare,
            .initialLayout = vk::ImageLayout::eUndefined,
            .finalLayout = vk::ImageLayout::eTransferSrcOptimal,
        };

        vk::AttachmentReference2 colorBlendAR {
            .attachment = 0,
            .layout = vk::ImageLayout::eColorAttachmentOptimal,
        };

        vk::AttachmentDescription2 depthAD {
            .format = vk.getDepthFormat(),
            .samples = vk::SampleCountFlagBits::e1,
            .loadOp = vk::AttachmentLoadOp::eClear,
            .storeOp = vk::AttachmentStoreOp::eDontCare,
            .stencilLoadOp = vk::AttachmentLoadOp::eDontCare,
            .stencilStoreOp = vk::AttachmentStoreOp::eDontCare,
            .initialLayout = vk::ImageLayout::eUndefined,
            .finalLayout = vk::ImageLayout::eDepthStencilAttachmentOptimal,
        };

        vk::AttachmentReference2 depthAR {
            .attachment = 1,
            .layout = vk::ImageLayout::eDepthStencilAttachmentOptimal,
        };

        vk::SubpassDescription2 subpass {
            .pipelineBindPoint = vk::PipelineBindPoint::eGraphics,
            .colorAttachmentCount = 1,
            .pColorAttachments = &colorBlendAR,
            .pDepthStencilAttachment = &depthAR,
        };

        // wait until image is available
        vk::SubpassDependency2 colorAttachmentSD {
            .srcSubpass = VK_SUBPASS_EXTERNAL,
            .dstSubpass = 0,
            .srcStageMask = vk::PipelineStageFlagBits::eColorAttachmentOutput | vk::PipelineStageFlagBits::eEarlyFragmentTests,
            .dstStageMask = vk::PipelineStageFlagBits::eColorAttachmentOutput | vk::PipelineStageFlagBits::eEarlyFragmentTests,
            .srcAccessMask = NO_ACC_FLAGS,
            .dstAccessMask = vk::AccessFlagBits::eColorAttachmentWrite | vk::AccessFlagBits::eDepthStencilAttachmentWrite,
        };

        // image layout transition to transfer dst
        vk::SubpassDependency2 layoutTransitionSD {
            .srcSubpass = 0,
            .dstSubpass = VK_SUBPASS_EXTERNAL,
            .srcStageMask = vk::PipelineStageFlagBits::eColorAttachmentOutput,
            .dstStageMask = vk::PipelineStageFlagBits::eTransfer,
            .srcAccessMask = vk::AccessFlagBits::eColorAttachmentWrite,
            .dstAccessMask = vk::AccessFlagBits::eTransferRead,
            .dependencyFlags = vk::DependencyFlagBits::eByRegion,
        };

        /* vk::SubpassDependency2 dependency { */
        /* dependency.srcSubpass = VK_SUBPASS_EXTERNAL; */
        /* dependency.dstSubpass = 0; */
        /* dependency.srcStageMask = vk::PipelineStageFlagBits::eColorAttachmentOutput | vk::PipelineStageFlagBits::eEarlyFragmentTests; */
        /* dependency.srcAccessMask = 0; */
        /* dependency.dstStageMask = vk::PipelineStageFlagBits::eColorAttachmentOutput | vk::PipelineStageFlagBits::eEarlyFragmentTests; */ 
        /* dependency.dstAccessMask = vk::AccessFlagBits::eColorAttachmentWrite | vk::AccessFlagBits::eDepthStencilAttachmentWrite; */

        std::vector<vk::AttachmentDescription2> attachments = { colorBlendAD, depthAD, };
        std::vector<vk::SubpassDependency2> dependencies = { colorAttachmentSD, layoutTransitionSD };
        vk::RenderPassCreateInfo2 renderPassCI {
        	/* .attachmentCount = static_cast<uint32_t>(attachments.size()), */
        	/* .pAttachments = attachments.data(), */
        	/* .subpassCount = 1, */
        	/* .pSubpasses = &subpass, */
        	/* .dependencyCount = static_cast<uint32_t>(dependencies.size()), */
        	/* .pDependencies = dependencies.data(), */
            /* .dependencyCount = 0, */
            /* .pDependencies = nullptr, */
            /* .correlatedViewMaskCount = 0, */
            /* .pCorrelatedViewMasks = nullptr, */
        };
        renderPassCI.setAttachments(attachments);
        renderPassCI.setDependencies(dependencies);
        renderPassCI.setSubpasses(subpass);
        vk::Result result = vk.getDevice().createRenderPass2(&renderPassCI, nullptr, &renderPass); 
        if (result != vk::Result::eSuccess) {
            throw getVkException(result, "Could not create render pass", "Renderer3D::createRenderPass");
        }
        rLog.log0("createRenderPass: Created render pass.");
    }



//
// DESCRIPTORS
//
    void Renderer3D::createDescriptorResources() {
        rLog("createDescriptorResources: pool", reinterpret_cast<uint64_t>(static_cast<VkDescriptorPool>(descriptorPool)), "vs hpp:", reinterpret_cast<uint64_t>(&(*descriptorPool)));
        // LAYOUTS
        std::vector<vk::DescriptorSetLayoutBinding> uboDSLB;
        // SET 1: uniform buffer object
        uboDSLB.emplace_back(vk::DescriptorSetLayoutBinding {
            .binding = 0,
            .descriptorType = vk::DescriptorType::eUniformBuffer,
            .descriptorCount = 1,
            .stageFlags = vk::ShaderStageFlagBits::eVertex,
            /* .pImmutableSamplers = nullptr, */
        });

        /* // SET 1: combined image sampler */
        /* set1bindings.emplace_back(vk::DescriptorSetLayoutBinding { */
        /*     .binding = 1, */
        /*     .descriptorType = vk::DescriptorType::eCombinedImageSampler, */
        /*     .descriptorCount = 1, */
        /*     .stageFlags = vk::ShaderStageFlagBits::eFragment, */
        /*     /1* .pImmutableSamplers = nullptr, *1/ */
        /* }); */
        vk.createDescriptorSetLayout(uboDSLB, uboDSL);

#ifndef RENDERER_3D_RENDER_TO_SWAP_CHAIN
        // SET 2: combined image sampler
        std::vector<vk::DescriptorSetLayoutBinding> sampleFromRenderImageDSLB;
        sampleFromRenderImageDSLB.emplace_back(vk::DescriptorSetLayoutBinding {
            .binding = 0,
            .descriptorType = vk::DescriptorType::eCombinedImageSampler,
            .descriptorCount = 1,
            .stageFlags = vk::ShaderStageFlagBits::eFragment,
        });
        vk.createDescriptorSetLayout(sampleFromRenderImageDSLB, sampleFromRenderImageDSL);
#endif


        // POOL
        std::vector<vk::DescriptorPoolSize> poolSizes;
        poolSizes = { 
            { .type = vk::DescriptorType::eUniformBuffer, .descriptorCount = vk.getMaxFramesInFlight() },
#ifndef RENDERER_3D_RENDER_TO_SWAP_CHAIN
            { .type = vk::DescriptorType::eCombinedImageSampler, .descriptorCount = vk.getMaxFramesInFlight() },
#endif
        };
        vk.createDescriptorPool(poolSizes, vk.getMaxFramesInFlight(), descriptorPool);

        // SETS
        // SET 1
        std::vector<vk::DescriptorSetLayout> layouts1(vk.getMaxFramesInFlight(), uboDSL);
        vk.createDescriptorSets(layouts1, descriptorPool, uboDS);

        for (size_t i = 0; i < vk.getMaxFramesInFlight(); i++) {
            vk::DescriptorBufferInfo bufferI {
            	.buffer = uniformBuffers[i],
            	.offset = 0,
            	.range = VK_WHOLE_SIZE,  // sizeof(UniformBufferObject),
            };

            vk::WriteDescriptorSet descriptorW{
                .dstSet = uboDS[i],
                .dstBinding = 0,
                .dstArrayElement = 0,
                .descriptorCount = 1,
                .descriptorType = vk::DescriptorType::eUniformBuffer,
                .pBufferInfo = &bufferI,
            };
            /* descriptorW[0].pTexelBufferView = nullptr; */
            
            // write 1, copy 0
            vk.getDevice().updateDescriptorSets(descriptorW, nullptr);
        }  // for
        rLog.log0("createDescriptorResources: Created descriptor layout and sets for uniform buffer");

#ifndef RENDERER_3D_RENDER_TO_SWAP_CHAIN
        // SET 2:
        std::vector<vk::DescriptorSetLayout> layouts2(vk.getMaxFramesInFlight(), sampleFromRenderImageDSL);
        vk.createDescriptorSets(layouts2, descriptorPool, sampleFromRenderImageDS);

        for (size_t i = 0; i < vk.getMaxFramesInFlight(); i++) {
            vk::DescriptorImageInfo imageI {
            	.sampler = imageSampler,
            	.imageView = renderImageViews[i],
            	.imageLayout = vk::ImageLayout::eShaderReadOnlyOptimal,
            };

            vk::WriteDescriptorSet descriptorW{
                .dstSet = sampleFromRenderImageDS[i],
                .dstBinding = 0,
                .dstArrayElement = 0,
                .descriptorCount = 1,
                .descriptorType = vk::DescriptorType::eCombinedImageSampler,
                .pImageInfo = &imageI,
            };
            // write 1, copy 0
            vk.getDevice().updateDescriptorSets(descriptorW, nullptr);
        } // for
        rLog.log0("createDescriptorResources: Created descriptor layout and sets for sampling from renderImages");
#endif
    }

//
// MODEL
//
	void Renderer3D::loadModel() {
        // load model into VerticesAndIndices struct
        rLog.log1("Renderer3D: loadModel: loading model");
        vk.loadModel(settings::MODEL_PATH, model);
        // TODO use correct type
        vk::DeviceSize requiredVertexBufferSize = model.vertices.size() * sizeof(Vertex3D);
        vk::DeviceSize requiredIndexBufferSize = model.indices.size() * sizeof(uint32_t);

        if (requiredVertexBufferSize > vertexBufferSize) { throw VkException("Renderer3D::loadModel: vertex buffer too small"); }
        if (requiredIndexBufferSize > indexBufferSize) { throw VkException("Renderer3D::loadModel: index buffer too small"); }

        rLog.log0("Renderer3D: loadModel: filling vertex buffer");
        // copy to vertexBuffer
        // create staging buffer
        vk::Buffer stagingBuffer;
        MemoryInfo stagingBufferMemory;
        vk.createBuffer(requiredVertexBufferSize, vk::BufferUsageFlagBits::eTransferSrc, vk::MemoryPropertyFlagBits::eHostVisible | vk::MemoryPropertyFlagBits::eHostCoherent, stagingBuffer, stagingBufferMemory);
        // fill staging buffer
        void* data;
        vk::Result result = vk.getDevice().mapMemory(stagingBufferMemory.memory, stagingBufferMemory.offset, requiredVertexBufferSize, NO_MEM_FLAGS, &data);
        if (result != vk::Result::eSuccess) {
            throw getVkException(result, "Failed to map staging buffer for vertices", "Renderer3D::loadModel");
        }

        memcpy(data, model.vertices.data(), requiredVertexBufferSize);
        vk.getDevice().unmapMemory(stagingBufferMemory.memory);
        // fill vertex buffer 
        vk.copyBuffer(stagingBuffer, vertexBuffer, requiredVertexBufferSize);
        vk.destroyBuffer(stagingBuffer, stagingBufferMemory);

        rLog.log0("Renderer3D: loadModel: filling index buffer");
        data = nullptr;
        // copy to index buffer
        vk.createBuffer(requiredIndexBufferSize, vk::BufferUsageFlagBits::eTransferSrc, vk::MemoryPropertyFlagBits::eHostVisible | vk::MemoryPropertyFlagBits::eHostCoherent, stagingBuffer, stagingBufferMemory);
        // fill staging buffer
        result = vk.getDevice().mapMemory(stagingBufferMemory.memory, stagingBufferMemory.offset, requiredIndexBufferSize, NO_MEM_FLAGS, &data);
        if (result != vk::Result::eSuccess) {
            throw getVkException(result, "Failed to map staging buffer for indices", "Renderer3D::loadModel");
        }
        memcpy(data, model.indices.data(), requiredIndexBufferSize);
        vk.getDevice().unmapMemory(stagingBufferMemory.memory);
        // fill index buffer 
        vk.copyBuffer(stagingBuffer, indexBuffer, requiredIndexBufferSize);
        vk.destroyBuffer(stagingBuffer, stagingBufferMemory);
	}

//
// RENDERING
//
    void Renderer3D::recordCommandBuffer(uint32_t imageIndex, uint32_t currentFrame) {
        vk::CommandBufferBeginInfo commandBufferBI;
        vk::Result result = commandBuffers[currentFrame].begin(commandBufferBI); 
        if (result != vk::Result::eSuccess) {
            throw getVkException(result, "Failed to begin 3D command buffer", "Renderer3D::recordCommandBuffer");
        }

        // clear
        std::array<vk::ClearValue, 2> clearValues{};
        vk::ClearColorValue clearColor{};
        clearColor.setFloat32({0.1f, 0.0f, 0.0f, 0.5f});
        clearValues[0].setColor(std::move(clearColor));
        clearValues[1].setDepthStencil({1.0f, 0});

        vk::RenderPassBeginInfo renderPassBI {
        	.renderPass = renderPass,
        	.framebuffer = framebuffers[imageIndex],
            .renderArea { .offset = { 0, 0 }, .extent = vk.getScExtent() }
        };
        renderPassBI.setClearValues(clearValues);

        commandBuffers[currentFrame].beginRenderPass(renderPassBI, vk::SubpassContents::eInline);

        commandBuffers[currentFrame].bindPipeline(vk::PipelineBindPoint::eGraphics, pipelines[R3Render].pipeline);
        vk::Buffer vertexBuffers[] = { vertexBuffer };
        vk::DeviceSize offsets[] = {0};
        uint32_t bindingCount = 1;
        uint32_t firstBinding = 0;
        commandBuffers[currentFrame].bindVertexBuffers(firstBinding, bindingCount, vertexBuffers, offsets);
        // TODO use correct index type!
        commandBuffers[currentFrame].bindIndexBuffer(indexBuffer, NO_OFFSET, vk::IndexType::eUint32);

        uint32_t firstSet = 0;
        std::vector<vk::DescriptorSet> descriptorSets = { uboDS[currentFrame], textureManager.getDescriptorSet() };
        commandBuffers[currentFrame].bindDescriptorSets(vk::PipelineBindPoint::eGraphics, pipelines[R3Render].layout, firstSet, descriptorSets, {});

        int instanceCount = 1;
        int firstIndex = 0;
        int firstInstance = 0;
        commandBuffers[currentFrame].drawIndexed(static_cast<uint32_t>(model.indices.size()), instanceCount, firstIndex, NO_OFFSET, firstInstance);

        /* commandBuffers[currentFrame].bindPipeline(vk::PipelineBindPoint::eGraphics, pipelines[blendToSC].pipeline); */
        /* commandBuffers[currentFrame].bindDescriptorSets(vk::PipelineBindPoint::eGraphics, pipelines[blendToSC].layout, firstSet, sampleFromRenderImageDS[currentFrame], {}); */
        /* // draw hardcoded fullscreen quad */
        /* commandBuffers[currentFrame].draw(6, 1, 0, 0); */
        commandBuffers[currentFrame].endRenderPass();

        vk.copyImageToImage(commandBuffers[currentFrame], renderImages[imageIndex], vk.getScImages()[imageIndex], vk.getScExtent());
        result = commandBuffers[currentFrame].end();
        if (result != vk::Result::eSuccess) {
            rLog.error("Failed to record 3D - command buffer", "VkResult:", result);
            throw getVkException(result, "Failed to record 3D - command buffer", "Renderer3D::recordCommandBuffer");
        }
        vk.submitThisFrame(commandBuffers[currentFrame]);
    }
//
// RENDERING
//
    void Renderer3D::drawFrame(uint32_t imageIndex) {
        commandBuffers[vk.getCurrentFrame()].reset(NO_CMD_RESET_FLAGS);
        /* recordCommandBuffer(imageIndex, vk.getCurrentFrame()); */
        updateUniformBuffer();
        recordCommandBuffer(imageIndex, vk.getCurrentFrame());
    }


    void Renderer3D::createUniformBuffers() {
        vk::DeviceSize bufferSize = sizeof(UniformBufferObject);

        uniformBuffers.resize(vk.getMaxFramesInFlight());
        uniformBuffersMemory.resize(vk.getMaxFramesInFlight());

        for (size_t i = 0; i < vk.getMaxFramesInFlight(); i++) {
            vk.createBuffer(bufferSize, vk::BufferUsageFlagBits::eUniformBuffer, vk::MemoryPropertyFlagBits::eHostVisible | vk::MemoryPropertyFlagBits::eHostCoherent, uniformBuffers[i], uniformBuffersMemory[i]);
        }
    }


    void Renderer3D::updateUniformBuffer() {
        static auto startTime = std::chrono::high_resolution_clock::now();
        auto currentTime = std::chrono::high_resolution_clock::now();
        float time = std::chrono::duration<float, std::chrono::seconds::period>(currentTime - startTime).count();

        // TODO use push constant instead of ubo
        UniformBufferObject ubo{};
        ubo.model = glm::rotate(glm::mat4(1.0f), time * std::numbers::pi_v<float> / 2, glm::vec3(0.0f, 0.0f, 1.0f));
        ubo.view = glm::lookAt(glm::vec3(2.0f, 2.0f, 2.0f), glm::vec3(0.0f, 0.0f, 0.0f), glm::vec3(0.0f, 0.0f, 1.0f));
        ubo.projection = glm::perspective(glm::radians(45.0f), static_cast<float>(vk.getScExtent().width) / vk.getScExtent().height, 1.0f, 10.0f);
        /* ubo.model = glm::mat4(1); */
        /* ubo.view = glm::mat4(1); */
        /* ubo.projection = glm::mat4(1); */
        /* ubo.projection[1][1] *= -1;  // y coordinate inverted in opengl */
        MemoryInfo& uniformBufferMI = uniformBuffersMemory[vk.getCurrentFrame()];
        void* data;
        vk::Result result = vk.getDevice().mapMemory(uniformBufferMI.memory, uniformBufferMI.offset, sizeof(ubo), NO_MEM_FLAGS, &data);
        if (result != vk::Result::eSuccess) {
            throw getVkException(result, "Failed to map staging buffer", "Renderer3D::updateUniformBuffer");
        }
        memcpy(data, &ubo, sizeof(ubo));
        vk.getDevice().unmapMemory(uniformBufferMI.memory);
    }

}  // namespace gz::vlk
